# Copyright 2021 Google LLC
# Copyright 2021 Fraunhofer FKIE
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Runs coverage_test with PyInstaller to ensure coverage still functions."""

import os
import subprocess
import sys
import unittest


class PyInstallerCoverageTest(unittest.TestCase):

  def test_build_and_run(self):
    coverage_test = os.path.join(
        os.path.dirname(os.path.realpath(__file__)), "coverage_test.py")
    subprocess.check_call([
        sys.executable, "-m", "PyInstaller", "--distpath", "/tmp/pyinstaller",
        "--onefile", "--name", "coverage_test.pkg", coverage_test
    ])
    subprocess.check_call(["/tmp/pyinstaller/coverage_test.pkg"])


if __name__ == "__main__":
  unittest.main()
